/* $Workfile: OtxRaw.h $   $Revision: 13 $    $Modtime: 12/21/05 12:12p $  */
#ifndef OTX_RAWDATA_H
#define OTX_RAWDATA_H
/*******************************************************************************

  @doc PublicHwApi OtxRawData

  Copyright (C) 1999 - 2003 Odin TeleSystems Inc., All Rights Reserved

  OTX Generic Hardware Driver API

  @module
  OtxRawData | Prototypes for the Logical HDLC Device (Sender/Receiver) Functions

  @contents2
  OtxRaw.h |

  @index func | OtxRawData

  @contents2
  OTX Logical RawData Sender/Receiver Functions - OtxRawDataXXX() | The API functions for the Logical High-level Data Link Controller (HDLC) device

  @head3 OtxRawDataXXX() Function Reference |

  @index func | OtxRawData

*******************************************************************************/

#ifdef __cplusplus
  extern "C" {
#endif

#include "OtxType.h"
#include "OtxDef.h"
#include "OtxDataD.h"
#include "OtxToneD.h"

#include "OtxRawD.h"


/*******************************************************************************

  @func
    Initialize the Raw Data Sender.

  @comm
    Start Raw Data Sender operation with by calling the function <f OtxDrvEnable> and stop
    it by calling the function <f OtxDrvDisable>

    @ex Using the Raw Data Sender with a 16 kpbs (two bits) subchannel with 
    bLsbSentFirst==OTX_TRUE: When working with a 16 kbps subchannel the Raw Data Sender will extract
    2 bits from each octet passed to the device and put these bits in 4 consecutive frames. The 2 bits are 
    extracted from the octet according to the figure below: |


     |
     |7 6 5 4 3 2 1 0
     |---+---+---+---+
     |x x|x x|x x|x x|  Octet passed to the Raw Data Sender
     |---------------+
     | ^           ^
     | |           |
     | |           +----- These two bits are sent first
     | |
     | |
     | +----- These two bits are sent last

  @ex |
    // create logical devices
    {
      OTX_RESULT result;
      OTX_INT32 i;
      OTX_UINT32 nIrpQueueLimit;
      OTX_UINT32 nFifoSize;
      OTX_UINT32 iDevNo;
      OTX_UINT32 nLi = 2;

      for (i = 0; i < nLi; i++) {

        // hDspDevNo is the dsp device
        iDevNo = i;

        // Create a notification Event and Event Queue for the Raw Data Sender
        result = OtxDrvCreateEvent(&g_hLogicDevs.mRawDataSender[i].hNotificationEvent);
        result = OtxDrvCreateEventQueue(
                    128, 
                    g_hLogicDevs.mRawDataSender[i].hNotificationEvent, 
                    &g_hLogicDevs.mRawDataSender[i].hEventQueue); 
        result = OtxDrvEnable(
                    g_hLogicDevs.mRawDataSender[i].hEventQueue, 
                    OTX_TASK_SYNC, 
                    5000L); 

        // Create Raw Data Sender
        result = OtxDrvCreateLogicalDevice(
                    g_hPhysDevs.hDsp[iDevNo], 
                    OTX_LDEVICE_DATA_RAW_SENDER, 
                    g_hLogicDevs.mRawDataSender[i].hEventQueue, 
                    i, 
                    &g_hLogicDevs.mRawDataSender[i].hHandle); 

        // Store sourceId so that we can easily print it from the Sender Thread
        g_hLogicDevs.mRawDataSender[i].nSourceId = i;   

        // Increase the IRP Queue limit for fast pipes (to be able to buffer more transmit data in the driver)
        result = OtxDrvGetAttributeValueUint32(
                    g_hLogicDevs.mRawDataSender[i].hHandle, 
                    OTX_ATTR_DEVICE_IRP_QUEUE_LIMIT, 
                    &nIrpQueueLimit); 

        // create a notification Event and Event Queue for the Raw Data Receiver

        result = OtxDrvCreateEvent(&g_hLogicDevs.mRawDataReceiver[i].hNotificationEvent); 
        result = OtxDrvCreateEventQueue(
                    128, 
                    g_hLogicDevs.mRawDataReceiver[i].hNotificationEvent, 
                    &g_hLogicDevs.mRawDataReceiver[i].hEventQueue); 
        result = OtxDrvEnable(
                    g_hLogicDevs.mRawDataReceiver[i].hEventQueue, 
                    OTX_TASK_SYNC, 
                    5000L); 

        // Create Raw Data Receiver
        result = OtxDrvCreateLogicalDevice(
                    g_hPhysDevs.hDsp[iDevNo], 
                    OTX_LDEVICE_DATA_RAW_RECEIVER, 
                    g_hLogicDevs.mRawDataReceiver[i].hEventQueue, 
                    i, 
                    &g_hLogicDevs.mRawDataReceiver[i].hHandle); 
        
        // store sourceId so that we can easily print it from the Receiver Thread
        g_hLogicDevs.mRawDataReceiver[i].nSourceId = i; 

  
        // Modify the FIFO size of the Raw Data Receiver
        result = OtxDrvGetAttributeValueUint32(
                    g_hLogicDevs.mRawDataReceiver[i].hHandle, 
                    OTX_ATTR_RAWDATA_RECEIVER_FIFO_SIZE, 
                    &nFifoSize); 

        // Prevent the driver from generating the OTX_RAWDATA_RECEIVER_EC_DATA_AVAILABLE event 
        // (to reduce driver overhead)
        result = OtxDrvSetAttributeValueBool(
                    g_hLogicDevs.mRawDataReceiver[i].hHandle, 
                    OTX_ATTR_RAWDATA_RECEIVER_GENERATE_DATA_EVENTS, 
                    OTX_FALSE); 

      }

    }


    // connect devices
    {
      OTX_INT32 nLi, nDsp;
      OTX_RESULT result;

      OTX_UINT8 sqTimeSlotMask[32];

      // create the Pipes (one for each direction) between the two DSPs and the two LIs
      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        OTX_INT32 nCapacity;
        OTX_INT32 nTs;

        // create Raw Data pipes (one for each direction) Use all timeslots except TS0.
        memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
  
        // Standard (64kpbs) or superchannel
        nCapacity = 64;

        // creating pipe for RawData Sender/Receiver 

        for (nTs = g_demoEnv.nStartingTimeslot; nTs < 32; nTs++) {
      
          if (nTs == 0) {
            continue; // TS0 is sync timeslot
          }
      
          sqTimeSlotMask[nTs] = 0xFF; 
          nCapacity += 64;
        }
    
        result = OtxDrvCreatePipe(
                    nCapacity, 
                    &g_hPipes[nLi].hLiDspRawData); 
        result = OtxDrvCreatePipe(
                    nCapacity, &g_hPipes[nLi].hDspLiRawData); 
  
        // connect pipe: Dsp -> Li 
        result = OtxDrvConnectPipe(
                    g_hPipes[nLi].hDspLiRawData, 
                    OTX_PIPE_INPUT,  
                    g_hPhysDevs.hDsp[nLi], 
                    -1, 
                    sqTimeSlotMask, 
                    32); 
        result = OtxDrvConnectPipe(
                    g_hPipes[nLi].hDspLiRawData, 
                    OTX_PIPE_OUTPUT, 
                    g_hPhysDevs.hLi[nLi],  
                    -1, 
                    sqTimeSlotMask, 
                    32);

        // Connect pipe: Li  -> Dsp 
        result = OtxDrvConnectPipe(
                    g_hPipes[nLi].hLiDspRawData, 
                    OTX_PIPE_INPUT,  
                    g_hPhysDevs.hLi[nLi],  
                    -1, 
                    sqTimeSlotMask, 
                    32); 
        result = OtxDrvConnectPipe(
                    g_hPipes[nLi].hLiDspRawData, 
                    OTX_PIPE_OUTPUT, 
                    g_hPhysDevs.hDsp[nLi], 
                    -1, 
                    sqTimeSlotMask, 
                    32); 
      }    

      // connect logical devices within DSP
      for (nDsp = 0; nDsp < 2; nDsp++) {
        // connecting RawData Sender/Receiver Logical Device 
        // Receive direction:
        // Connect Raw Data Receiver
        result = OtxDrvConnectLogicalDevice(
                    g_hPipes[nDsp].hLiDspRawData,  
                    -1,  
                    g_hLogicDevs.mRawDataReceiver[nDsp].hHandle, 
                    -1); 
    
        // Transmit direction:
        // Connect Raw Data Sender
        result = OtxDrvConnectLogicalDevice(
                    g_hLogicDevs.mRawDataSender[nDsp].hHandle, 
                    -1, 
                    g_hPipes[nDsp].hDspLiRawData, 
                    -1);
      }

      // initialize logical devices
      {
        OTX_RESULT result;
        OTX_INT32 nDsp;

        for (nDsp = 0; nDsp < 2; nDsp++) {
    
          // initialize the Raw Data Receivers (hosted in the DSP)
          result = OtxRawDataReceiverInit(
                      g_hLogicDevs.mRawDataReceiver[nDsp].hHandle, 
                      OTX_DATA_FORMAT_UINT8,  
                      0xFF,      
                      OTX_FALSE, 
                      OTX_TRUE); 

          // initialize the Raw Data Senders (hosted in the DSP)
          result = OtxRawDataSenderInit(
                      g_hLogicDevs.mRawDataSender[nDsp].hHandle, 
                      OTX_DATA_FORMAT_UINT8,  
                      0xFF,     
                      OTX_TRUE);
        }
    }

    // enable logical devices
    for (nDsp = 0; nDsp < 2; nDsp++) {
      // enable Raw Data Sender
      result = OtxDrvEnable(g_hLogicDevs.mRawDataSender[nDsp].hHandle, 
                  OTX_TASK_SYNC, 
                  5000L); 
    }
    
    // enable Raw Data Receivers in the thread, see example on
    @ex <f OtxRawDataReadData> |    

  @xref
    <f OtxRawDataSenderInitEx>
    <f OtxRawDataReceiverInit>
    <f OtxRawDataSendData>
    <e OtxLogicalDeviceTypeE.OTX_LDEVICE_DATA_RAW_SENDER>
    <t OtxDataFormatE>
 
*******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxRawDataSenderInit(
  IN OTX_HANDLE     hRawDataDevice,   // @parm Handle to logical Raw Data Device
  IN OtxDataFormatE eDataFormat,      // @parm Format of the Raw Data supplied to this device. 
                                      // Use <e OtxDataFormatE.OTX_DATA_FORMAT_UINT8> if no conversion to be applied. 
                                      // Use <e OtxDataFormatE.OTX_DATA_FORMAT_COMPAND8_A_LAW> or <e OtxDataFormatE.OTX_DATA_FORMAT_COMPAND8_U_LAW> 
                                      // to supply 16-bit signed integer values
                                      // in the calls to <f OtxRawDataSendData> and let the device compress these value to 8-bit
                                      // A-law/U-law values before they are placed in the outgoing data stream.
                                      // <nl>Note that maximim range for A-law and U-law is a 14-bit value. Therefore the maximum signed value
                                      // supplied to this function in A-law/U-law format is in the range of -8192 <lt>=x  <lt> 8192.
                                      // If the value is larger than this range please use the <f OtxRawDataSenderInitEx> function
                                      // instead to initialize this device and supply the <e OtxToneGainE.OTX_TONE_GAIN_DG_MINUS_6DB> as the gain
                                      // parameter.
  IN OTX_UINT8      flag,             // @parm Value to send when there is no more Raw Data to send
  IN OTX_BOOL       bLsbSentFirst     // @parm Only applicable for subchannels (pipe capacity of 8, 16, or 32kpbs). Determines in which order bits
                                      // are extracted from the each octet before they are placed into a subchannel pipe. 
                                      // If TRUE the device will send the least significant bits of the octet first, 
                                      // and send the most significant bits last. 
                                      // If FALSE the device will send the most significant bits of the octet first, 
                                      // and send the least significant bits last. 
);


/*******************************************************************************

  @func
    Extended Initialization function for the Raw Data Sender.It takes one additional
    parameter in addition to the parameters that <f OtxRawDataSenderInit> takes. Either
    <f OtxRawDataSenderInit> or this function should be called for the Raw Data Sender
    device before it can be enabled.

  @comm
    Start Raw Data Sender operation with by calling the function <f OtxDrvEnable> and stop
    it by calling the function <f OtxDrvDisable>

  @xref
    <f OtxRawDataSenderInit>
    <f OtxRawDataReceiverInit>
    <f OtxRawDataSendData>
    <e OtxLogicalDeviceTypeE.OTX_LDEVICE_DATA_RAW_SENDER>
    <t OtxDataFormatE>
 
*******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxRawDataSenderInitEx(
  IN OTX_HANDLE     hRawDataDevice,   // @parm Handle to logical Raw Data Device
  IN OtxDataFormatE eDataFormat,      // @parm Format of the Raw Data supplied to this device. 
                                      // Use <e OtxDataFormatE.OTX_DATA_FORMAT_UINT8> if no conversion to be applied. 
                                      // Use <e OtxDataFormatE.OTX_DATA_FORMAT_COMPAND8_A_LAW> or <e OtxDataFormatE.OTX_DATA_FORMAT_COMPAND8_U_LAW> 
                                      // to supply 16-bit signed integer values
                                      // in the calls to <f OtxRawDataSendData> and let the device compress these value to 8-bit
                                      // A-law/U-law values before they are placed in the outgoing data stream.
  IN OtxToneGainE   eToneGain,        // @parm Normally the <e OtxToneGainE.OTX_TONE_GAIN_0DB> would be used. However, for a data format of 
                                      // <e OtxDataFormatE.OTX_DATA_FORMAT_COMPAND8_A_LAW> or <e OtxDataFormatE.OTX_DATA_FORMAT_COMPAND8_U_LAW>
                                      // other values can be used. These two formats have range of 14 bits (which are provided to the device
                                      // as signed 16-bit values). If the values are normalized in the most significant 14 bits of the 16-bit word
                                      // then the value of <e OtxToneGainE.OTX_TONE_GAIN_DG_MINUS_6DB> should be used for this parameter. If the values
                                      // are normalized in the least significant 14 bits of the 16-bit word
                                      // then the value of <e OtxToneGainE.OTX_TONE_GAIN_0DB> should be used for this parameter. The <f OtxRawDataSenderInit>
                                      // function uses a hardcoded value of <e OtxToneGainE.OTX_TONE_GAIN_0DB>.
  IN OTX_UINT8      flag,             // @parm Value to send when there is no more Raw Data to send
  IN OTX_BOOL       bLsbSentFirst     // @parm Only applicable for subchannels (pipe capacity of 8, 16, or 32kpbs). Determines in which order bits
                                      // are extracted from the each octet before they are placed into a subchannel pipe. 
                                      // If TRUE the device will send the least significant bits of the octet first, 
                                      // and send the most significant bits last. 
                                      // If FALSE the device will send the most significant bits of the octet first, 
                                      // and send the least significant bits last. 
);


/*******************************************************************************

  @func
    Initializate the HDLC Receiver.

  @comm
    Start HDLC operation with by calling the function <f OtxDrvEnable> and stop
    it by calling the function <f OtxDrvDisable>

    @ex Using the Raw Data Receiver with a 16 kpbs (two bits) subchannel with 
    bLsbReceivedFirst==OTX_TRUE. When working with a 16 kbps subchannel the Raw Data Receiver will extract
    2 bits from 4 consecutive frames and format the 4*2=8 bits into one octet 
    according to the figure below: |
     
     |
     |7 6 5 4 3 2 1 0
     |---+---+---+---+
     |x x|x x|x x|x x|  Octet received by the Raw Data Receiver
     |---------------+
     | ^           ^
     | |           |
     | |           +----- These two bits are the first two bits received 
     | |                  from the subchannel pipe
     | |
     | |
     | +----- These two bits are the last two bits received 
     |        from the subchannel pipe

  see example on <f OtxRawDataSenderInit>

  @xref
    <f OtxRawDataSenderInit>
    <nl><f OtxRawDataReadData>
    <nl><e OtxLogicalDeviceTypeE.OTX_LDEVICE_DATA_RAW_RECEIVER>
    <nl><t OtxDataFormatE>

*******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxRawDataReceiverInit(
  IN OTX_HANDLE     hRawDataDevice,   // @parm Handle to logical Raw Data Device
  IN OtxDataFormatE eDataFormat,      // @parm Format of the received Raw Data. 
                                      // Use <e OtxDataFormatE.OTX_DATA_FORMAT_UINT8> if no data conversion to be applied. 
                                      // Use <e OtxDataFormatE.OTX_DATA_FORMAT_COMPAND8_A_LAW> or <e OtxDataFormatE.OTX_DATA_FORMAT_COMPAND8_U_LAW> 
                                      // to expand the 8-bit values from the incoming data stream to 16-bit signed integer values
                                      // retrieved in the <f OtxRawDataReadData> calls.

  IN OTX_UINT8      flag,             // @parm Flag to be extracted if <p bExtractFlag> is TRUE
  IN OTX_BOOL       bExtractFlag,     // @parm if TRUE then received octets that equal the flag value are extracted from the receive data stream
  IN OTX_BOOL       bLsbReceivedFirst // @parm Only applicable for subchannels (pipe capacity of 8, 16, or 32kpbs). If TRUE then the received octets are 
                                      // formatted so that the least significant bits of the octet contains the the first received bits of 
                                      // the subchannel. If FALSE then the received octets are 
                                      // formatted so that the most significant bits of the octet contains the the first received bits of 
                                      // the subchannel;  
);


/*******************************************************************************

  @func
    Sends transparent data.

  @comm
    Usaully it is used in Raw Data Sender Thread

  @ex |
    
    {
      OtxEventDataS eventData;
      OTX_RESULT result;
      OTX_RESULT otxRc;
      char msgStr[256];
      clock_t startTime, stopTime;

      // Discard any events from  event queue (if any)
      do {
        result = OtxDrvGetEventData(
                    g_hLogicDevs.mRawDataSender.hEventQueue, 
                    &eventData); 
      } while (OTX_RESULT_CODE(result) == OTX_S_OK);
 
      // Flush (empty) the queue of raw data octets waiting to be transmitted (if any).
      result = OtxRawDataSenderFlush(g_hLogicDevs.mRawDataSender.hHandle); 

      startTime = clock();
      g_hLogicDevs.mRawDataSender.nTaskRef = 1;

      // Stuff the Raw Data Sender device full with data before we start to wait for OTX_RAWDATA_SENDER_EC_BUF_DEPLETED events
      do {
        g_hLogicDevs.mRawDataSender.cbTxBuf = fread(g_hLogicDevs.mRawDataSender.sqTxBuf, sizeof(OTX_UINT8), DEMO_RAW_TX_BUF_SIZE, g_hLogicDevs.mRawDataSender.pTxFile);
  
        otxRc = OtxRawDataSendData(
                   g_hLogicDevs.mRawDataSender.hHandle, 
                   g_hLogicDevs.mRawDataSender.sqTxBuf, 
                   g_hLogicDevs.mRawDataSender.cbTxBuf, 
                   g_hLogicDevs.mRawDataSender.nTaskRef, 
                   OTX_TIME_INFINITY);

        if (OTX_RESULT_CODE(otxRc) != OTX_E_BUSY) {
          g_hLogicDevs.mRawDataSender.nTaskRef++;
        }

      } while (g_hLogicDevs.mRawDataSender.cbTxBuf && (OTX_RESULT_CODE(otxRc) != OTX_E_BUSY));


      result = OtxDrvEnable(
                  g_hLogicDevs.mRawDataSender.hHandle,  
                  OTX_TASK_SYNC, 
                  5000L); 
  
      if (g_hLogicDevs.mRawDataSender.cbTxBuf) {
        DemoLogf("DemoRawDataSender: Primed the Sender with %d packets.\n", g_hLogicDevs.mRawDataSender.nTaskRef);
      } else {
        DemoLogf("DemoRawDataSender: The entire file fit in the driver buffer. Closing file\n");
        fclose(g_hLogicDevs.mRawDataSender.pTxFile);
        g_hLogicDevs.mRawDataSender.pTxFile = NULL;
      }
  
  
      // Wait for events from the Raw Data Sender
      while (1) {
        OtxDrvWaitForSingleEvent(
           g_hLogicDevs.mRawDataSender.hNotificationEvent, 
           3);
        
        do {
          result = OtxDrvGetEventData(
                      g_hLogicDevs.mRawDataSender.hEventQueue, 
                      &eventData); 

          if (OTX_RESULT_CODE(result) == OTX_S_OK) {
            // We have a driver event
            switch (eventData.m_eDeviceType) {
              case OTX_LDEVICE_DATA_RAW_SENDER:
              // Event from Raw Data Sender 

              switch (eventData.m_nCode) {
                case OTX_RAWDATA_SENDER_EC_BUF_DEPLETED:
                {
                  OTX_RESULT otxRc;
                
                  DemoLogf("Buffer sent (eventData.m_nRequestId = %d)\n", eventData.m_nRequestId);
                
                  if (g_hLogicDevs.mRawDataSender.pTxFile) {
                    while (g_hLogicDevs.mRawDataSender.cbTxBuf) {
                      otxRc = OtxRawDataSendData(
                                 g_hLogicDevs.mRawDataSender.hHandle, 
                                 g_hLogicDevs.mRawDataSender.sqTxBuf, 
                                 g_hLogicDevs.mRawDataSender.cbTxBuf, 
                                 g_hLogicDevs.mRawDataSender.nTaskRef, 
                                 OTX_TIME_INFINITY);
                
                      if (OTX_RESULT_CODE(otxRc) == OTX_E_BUSY) {
                        break;
                      } else {
                        g_hLogicDevs.mRawDataSender.nTaskRef++;
                      }
                
                      g_hLogicDevs.mRawDataSender.cbTxBuf = fread(g_hLogicDevs.mRawDataSender.sqTxBuf, sizeof(OTX_UINT8), DEMO_RAW_TX_BUF_SIZE, g_hLogicDevs.mRawDataSender.pTxFile);
                    
                    }
                  
                    if (!g_hLogicDevs.mRawDataSender.cbTxBuf) {
                      DemoLogf("DemoRawDataSender: EOF reached. Closing file\n");
                      fclose(g_hLogicDevs.mRawDataSender.pTxFile);
                      g_hLogicDevs.mRawDataSender.pTxFile = NULL;
                    }
                  }
                }
                break;
                
                case OTX_RAWDATA_SENDER_EC_COMPLETED:
                  DemoLogf("DemoRawDataSender: Raw Data transmission ended.\n");
                  goto END;
                  break;
                
                default:
                  otxRc = OtxDrvEventCode2String(
                             eventData.m_hDevice, 
                             eventData.m_nCode, 
                             256, 
                             &msgStr[0]);
                  // Unexpected event from this device
                  if (OTX_SUCCESS(otxRc)) {
                    DemoLogf("DemoRawDataSender: Unexpected Event: %s \n", msgStr);
                  } else {
                    DemoLogf("DemoRawDataSender: Invalid Event: %ld \n", eventData.m_nCode);
                  }
                break;
              }
              break;
            
              default:
                // Unhandled event
                DemoLogf("DemoRawDataSender: Event from an unknown object 0x%X\n", eventData.m_hDevice);
              break;
            }
          }
        } while (OTX_RESULT_CODE(result) == OTX_S_OK);
      }

    END:
  
      stopTime = clock();

      DemoLogf(".\n");

      result = OtxDrvDisable(
                  g_hLogicDevs.mRawDataSender.hHandle,  
                  OTX_TASK_SYNC, 
                  5000L); 
  
      DemoLogf("DemoRawDataSender: Done (time=%2.1f seconds, Last nTaskRef=%d).\n", (double)(stopTime - startTime) / CLOCKS_PER_SEC, g_hLogicDevs.mRawDataSender.nTaskRef-1);
    }

*******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxRawDataSendData(
  IN OTX_HANDLE     hRawDataDevice,   // @parm Handle to the Raw Data Sender Device
  IN OTX_UINT8      *sqData,          // @parm Data to be sent
  IN OTX_UINT32     nDataLen,         // @parm Length of the data to be send
  IN OTX_TASK_REF   nTaskRef,         // @parm See <t OTX_TASK_REF>
  IN OTX_TIME       nWaitMaxMs        // @parm See <t OTX_TIME>
);


/*******************************************************************************

  @func
    Flushes (empties) the queue of raw data octets waiting to be transmitted (if any).
  
  @ex |
    // discard any events from event queue (if any)
    do {
      result = OtxDrvGetEventData(
                  g_hLogicDevs.mRawDataSender.hEventQueue, 
                  &eventData ); 
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
 

    // empty the queue of raw data octets waiting to be transmitted (if any).
    result = OtxRawDataSenderFlush(g_hLogicDevs.mRawDataSender.hHandle); 

  @xref
    <f OtxRawDataSendData>

*******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxRawDataSenderFlush(
  IN OTX_HANDLE     hRawDataDevice    // @parm Handle to the Raw Data Sender Device
);


/*******************************************************************************

  @func
    Read data received by the Raw Data receiver

  @comm
    Usually it is used in Raw Data Receiver Thread 

  @rdesc Returns one of the following values:

    @flag OTX_S_OK | Raw Data Data was fetched
    @flag OTX_S_FALSE | No Raw Data Data available at this moment
    @flag OTX_E_BUFFER_TO_SMALL | The supplied user data buffer is too small to fit the received data

  @ex |

    // Raw Data Receiver Thread 
    {
      OtxEventDataS eventData;
      OTX_RESULT result;
      OTX_RESULT otxRc;
      OTX_UINT8  buf[2048];
      OTX_UINT32 nFrameLen;
      OTX_UINT32 nMaxFrameLen = 0;
      OTX_UINT32 nEventCount = 0;
      OTX_UINT32 nReadDataCount = 0;
      char msgStr[256];
      clock_t startTime, stopTime;

      startTime = clock();

      result = OtxDrvEnable(
                  g_hLogicDevs.mRawDataReceiver.hHandle,  
                  OTX_TASK_SYNC, 
                  5000L); 

      while (clock() < (startTime + g_hLogicDevs.mRawDataReceiver.nMsDuration)) {
        OtxDrvWaitForSingleEvent(
           g_hLogicDevs.mRawDataReceiver.hNotificationEvent, 
           OTX_TIME_INFINITY);
        do {
          result = OtxDrvGetEventData(
                      g_hLogicDevs.mRawDataReceiver.hEventQueue, 
                      &eventData); 

          nEventCount++;

          if (OTX_RESULT_CODE(result) == OTX_S_OK) {
            // We have a driver event
            switch (eventData.m_eDeviceType) {
              // Event from Raw Data Receiver 
              case OTX_LDEVICE_DATA_RAW_RECEIVER:

              switch (eventData.m_nCode) {
                case OTX_RAWDATA_RECEIVER_EC_DATA_AVAILABLE:
                {
                  OTX_BOOL bFrameEnd;
                  otxRc = OtxRawDataReadData(
                             eventData.m_hDevice, 
                             buf, 
                             2048, 
                             &nFrameLen, 
                             &bFrameEnd);
                  while (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {

                    nReadDataCount++;
                
                    if (g_hLogicDevs.mRawDataReceiver.pRxFile) {
                      int cbWritten;

                      if (nFrameLen > nMaxFrameLen) {
                      nMaxFrameLen = nFrameLen;
                      }

                      cbWritten = fwrite(buf, sizeof(OTX_UINT8), nFrameLen, g_hLogicDevs.mRawDataReceiver.pRxFile);

                      if (cbWritten != (int)nFrameLen) {
                        DemoLogf("RawDataReceiver[%d] Error: Could not write data to file.  (nFrameLen=%d, cbWritten=%d) \n", eventData.m_nSourceId, nFrameLen, cbWritten);
                      }
                  
                      if (bFrameEnd) {
                        DemoLogf("RawDataSender[%d] Frame End occurred (unexpectedly). Data reception complete. Closing data file. \n", eventData.m_nSourceId);
                        fclose(g_hLogicDevs.mRawDataReceiver.pRxFile);
                        g_hLogicDevs.mRawDataReceiver.pRxFile = NULL;
                        goto END;
                      }
                    }
                
                    otxRc = OtxRawDataReadData(
                               eventData.m_hDevice, 
                               buf, 
                               2048, 
                               &nFrameLen, 
                               &bFrameEnd);
                  }
              
                  if (OTX_RESULT_CODE(otxRc) != OTX_S_FALSE) {
                    DemoLogf("RawDataReceiver[%d] OtxRawDataReadData() was unsuccessful.\n", eventData.m_nSourceId);
                    DemoCheckRc(eventData.m_hDevice, otxRc);
                  }

                }
                break;
          
                default:
                  otxRc = OtxDrvEventCode2String(
                             eventData.m_hDevice, 
                             eventData.m_nCode, 
                             256, 
                             &msgStr[0]);
                  // Unexpected event from this device
                  if (OTX_SUCCESS(otxRc)) {
                    DemoLogf("RawDataReceiver[%d] Event: %s \n", eventData.m_nSourceId, msgStr);
                  } else {
                    DemoLogf("RawDataReceiver[%d] Invalid Event: %ld \n", eventData.m_nSourceId, eventData.m_nCode);
                  }
                break;
          
              }
              break;
            
              default:
              // Unhandled event
                DemoLogf("DemoRawDataSender: Event from an unknown object 0x%X\n", eventData.m_hDevice);
              break;
            }
        
          }
        } while (OTX_RESULT_CODE(result) == OTX_S_OK);

        OsaThreadSleep(256); // Give the driver some time to fill up more data for us to fetch (256ms fills 2048 bytes in a 64kbit/s pipe)
      }

      END:
  
        stopTime = clock();

        result = OtxDrvDisable(
                    g_hLogicDevs.mRawDataReceiver.hHandle,  
                    OTX_TASK_SYNC, 
                    5000L); 

        if (g_hLogicDevs.mRawDataReceiver.pRxFile) {
          fclose(g_hLogicDevs.mRawDataReceiver.pRxFile);
          g_hLogicDevs.mRawDataReceiver.pRxFile = NULL;
        }

        DemoLogf("DemoRawDataReceiver: nEventCount=%d, nReadDataCount=%d, nMaxFrameLen=%d\n", nEventCount, nReadDataCount, nMaxFrameLen);
        DemoLogf("DemoRawDataReceiver: Done (time=%2.1f seconds).\n", (double)(stopTime - startTime) / CLOCKS_PER_SEC);
    }

  @xref
    <f OtxRawDataReceiverInit>
    <nl><e OtxLogicalDeviceTypeE.OTX_LDEVICE_DATA_RAW_RECEIVER>
    <nl><f OtxRawDataSendData>

*******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxRawDataReadData(
  IN  OTX_HANDLE    hRawDataDevice,   // @parm Handle to the Raw Data Receiver Device
  OUT OTX_UINT8     *pDataBuf,        // @parm Data buffer for the received data
  IN  OTX_UINT32    nBufLen,          // @parm Length of the user data buffer
  OUT OTX_UINT32    *pnFrameLen,      // @parm Length of the received raw data buffer (number of octets)
  OUT OTX_BOOL      *pbFrameEnd       // @parm Set to OTX_TRUE if a flag (as specified in <f OtxRawDataReceiverInit>) followed the received data frame (valid if OTX_TRUE if <f OtxRawDataReceiverInit> was initialized with bExtractFlag=OTX_TRUE)
);


/*******************************************************************************

  @func
    Load an idle pattern for the Raw Data Sender devicie and begin transmission. The Idle Pattern
    is a sequence of data which will be send continuosly always when there is no
    actual user data to be send. User data can be send with <f OtxRawDataSendData> function
    on which case the user data frame in interleaved between two consequtive idle
    patterns. After the sending of user data has been completed,
    the Raw Data device will resume the sending of the idle pattern.
  
  @xref
   <f OtxRawDataStopIdlePattern>

  @todo

*******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxRawDataStartIdlePattern(
  IN OTX_HANDLE     hRawDataDevice,   // @parm Handle to the HDLC Device
  IN OTX_UINT8      *sqData,          // @parm Data to be sent
  IN OTX_UINT32     nDataLen          // @parm Length of the data to be send
);


/*******************************************************************************

  @func
    Stop sending of idle pattern on the specified Raw Data Sender device.
  
  @xref
   <f OtxRawDataStartIdlePattern>

  @todo

*******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxRawDataStopIdlePattern(
  IN OTX_HANDLE     hRawDataDevice    // @parm Handle to the HDLC Device
);



#ifdef __cplusplus
  } // extern "C"
#endif


/*******************************************************************************

  Copyright (C) Odin TeleSystems Inc., 1999 - 2003

  All rights reserved. No part of this computer program may be reproduced,
  stored in any retrieval system, or transmitted, in any form or by any
  means, electronic, mechanical, photocopying, recording, or otherwise,
  without the prior written permission of Odin TeleSystems Inc., or without
  a permission stipulated in the agreement/contract under which this computer
  program was supplied.

*******************************************************************************/
#endif // OTX_RAWDATA_H
